__author__ = "Blank Spruce"
__author_email__ = "blankspruce@protonmail.com"
__description__ = "A formatter to make your CMake code the real treasure"
__license__ = "MPL 2.0"
__title__ = "gersemi"
__url__ = "https://github.com/BlankSpruce/gersemi"
__version__ = "0.2.1"
